/*******************************************************************************
 * Copyright 2019 grondag
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License.  You may obtain a copy
 * of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
 * License for the specific language governing permissions and limitations under
 * the License.
 ******************************************************************************/
package grondag.xm.api.block;

import grondag.xm.api.modelstate.MutableModelState;
import grondag.xm.dispatch.XmBlockStateAccess;
import net.minecraft.class_1747;
import net.minecraft.class_1922;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2680;
import org.jetbrains.annotations.ApiStatus.Experimental;
import org.jetbrains.annotations.Nullable;

@Experimental
public interface XmBlockState {
	/**
	 * Minecraft block state associated with this Exotic Matter block state.
	 * Association is always 1:1.
	 */
	class_2680 blockState();

	/**
	 * If last parameter is false, does not perform a refresh from world for
	 * world-dependent state attributes. Use this option to prevent infinite
	 * recursion when need to reference some static state ) information in order to
	 * determine dynamic world state. Block tests are main use case for false.
	 */
	<T extends MutableModelState> T modelState(@Nullable class_1922 world, @Nullable class_2338 pos, boolean refreshFromWorld);

	default <T extends MutableModelState> T defaultModelState() {
		return modelState(null, null, false);
	}

	static @Nullable XmBlockState get(class_2680 fromState) {
		return ((XmBlockStateAccess)fromState).xm_toXmBlockState();
	}

	static @Nullable XmBlockState get(class_1747 fromItem) {
		return get(fromItem.method_7711());
	}

	static @Nullable XmBlockState get(class_2248 fromBlock) {
		return get(fromBlock.method_9564());
	}

	@SuppressWarnings("unchecked")
	static @Nullable <T extends MutableModelState> T modelState(class_2680 fromState, @Nullable class_1922 blockView, @Nullable class_2338 pos, boolean refresh) {
		final XmBlockState xmState = get(fromState);
		return xmState == null ? null : (T) xmState.modelState(blockView, pos, refresh);
	}
}
