/*******************************************************************************
 * Copyright 2019 grondag
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License.  You may obtain a copy
 * of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
 * License for the specific language governing permissions and limitations under
 * the License.
 ******************************************************************************/
package grondag.xm.api.primitive.simple;

import java.util.function.Function;
import net.minecraft.class_2350;
import org.jetbrains.annotations.ApiStatus.Experimental;
import grondag.fermion.orientation.api.OrientationType;
import grondag.xm.Xm;
import grondag.xm.api.mesh.WritableMesh;
import grondag.xm.api.mesh.XmMesh;
import grondag.xm.api.mesh.XmMeshes;
import grondag.xm.api.mesh.polygon.MutablePolygon;
import grondag.xm.api.modelstate.primitive.PrimitiveState;
import grondag.xm.api.paint.SurfaceTopology;
import grondag.xm.api.primitive.SimplePrimitive;
import grondag.xm.api.primitive.surface.XmSurface;
import grondag.xm.api.primitive.surface.XmSurfaceList;
import grondag.xm.api.texture.TextureOrientation;

@Experimental
public class Cube {
	private Cube() {}

	public static final XmSurfaceList SURFACES = XmSurfaceList.builder()
			.add("all", SurfaceTopology.CUBIC, XmSurface.FLAG_ALLOW_BORDERS)
			.build();

	public static final XmSurface SURFACE_ALL = SURFACES.get(0);

	static final Function<PrimitiveState, XmMesh> POLY_FACTORY = modelState -> {
		final WritableMesh mesh = XmMeshes.claimWritable();
		final MutablePolygon writer = mesh.writer();
		writer.colorAll(0, 0xFFFFFFFF);
		writer.lockUV(0, true);
		writer.rotation(0, TextureOrientation.IDENTITY);
		writer.sprite(0, "");
		writer.saveDefaults();

		writer.surface(SURFACE_ALL);
		writer.setupFaceQuad(class_2350.field_11033, 0, 0, 1, 1, 0, class_2350.field_11043);
		writer.append();

		writer.surface(SURFACE_ALL);
		writer.setupFaceQuad(class_2350.field_11036, 0, 0, 1, 1, 0, class_2350.field_11043);
		writer.append();

		writer.surface(SURFACE_ALL);
		writer.setupFaceQuad(class_2350.field_11034, 0, 0, 1, 1, 0, class_2350.field_11036);
		writer.append();

		writer.surface(SURFACE_ALL);
		writer.setupFaceQuad(class_2350.field_11039, 0, 0, 1, 1, 0, class_2350.field_11036);
		writer.append();

		writer.surface(SURFACE_ALL);
		writer.setupFaceQuad(class_2350.field_11043, 0, 0, 1, 1, 0, class_2350.field_11036);
		writer.append();

		writer.surface(SURFACE_ALL);
		writer.setupFaceQuad(class_2350.field_11035, 0, 0, 1, 1, 0, class_2350.field_11036);
		writer.append();

		return mesh.releaseToReader();
	};

	public static final SimplePrimitive INSTANCE = SimplePrimitive.builder()
			.surfaceList(SURFACES)
			.polyFactory(POLY_FACTORY)
			.orientationType(OrientationType.NONE)
			.build(Xm.id("cube"));
}
